/*
 This is a modified version of RES2IMP supplied by the original author.
 He has graciously permitted this code to enter the public domain
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <des/des.h>

#include "defs.h"

#ifdef _MSC_VER
#define strcasecmp strcmpi
#endif
typedef int (*rsrc_callback)(unsigned char[4],int, unsigned char*,int);
int found_keydata = 0;
unsigned char book_key[8];
unsigned char check_block[16];

int need_aes_key = 0;

int has_devicekey = 0;
unsigned char devicekey[8];

unsigned char * sb_header, * sb_dir_ptr;
int sb_header_len;
unsigned char * meta_data = 0;
int meta_len;
int num_res_files;
int processed_res_files;

FILE * fImp;

char * astrcat(char * a, char * b)
{
    int len;
    char * new;

    len = strlen(a) + strlen(b) + 1;
    new = (char *)malloc(len);
    if (new) 
    {
        strcpy(new,a);
        strcat(new,b);
    }
    return new;
}

int keycallback(unsigned char type[4],int id,unsigned char * data, int len)
{
    int ver;

    if (memcmp(type,"!!ky",4) != 0) return 0;
#if 0
    printf("Got a !!ky type of ID %d, %d bytes\n", id, len);
#endif
    ver = READ_BE_DWORD(data+8);
    if ((len != 36) || ((ver != 2) && (ver != 3)) )
    {
        fprintf(stderr,"Invalid !!ky data\n");
#if 0
        printf("---------\n");
        hexdump(data, len);
        printf("---------\n");
#endif
        return -1;
    }
    if (ver == 3) need_aes_key++;
    found_keydata++;
    memcpy(book_key, data+12, 8);
    memcpy(check_block, data+20, 16);
    return 0;
}

/* RSRC format:
 * +--------+-----------------+----+
 * |VERSION | 4 CHAR TYPE     | 00 |
 * +--------+-----------------+----+
 * |   00   | 4 INDEX OFFSET  |
 * +--------+-----------------+
 *
 *
 *  <---------------- 12 BYTES --------------------->
 * +---------+-------------+-------------+-----------+-- ... ---+
 * | RSRC ID | LENGTH      | OFFSET      | EXT COUNT | EXT DATA |
 * +---------+-------------+-------------+-----------+-- ... ---+
 *
 */
int handle_v1_file(unsigned char * fileptr, long filelen,rsrc_callback callback)
{
    unsigned long int offset;
    int ext_count,id, index_offset, data_len;
    int res, j;

    if (filelen < 24) 
    {
        fprintf(stderr,"File is too small -- %d bytes (minimum size is 24).\n",
            filelen);
        return 1; /* Too small for header+index*/
    }
    id = READ_BE_WORD(fileptr);
    if (id != 1)
    {
        /** fprintf(stderr,"Version is %d, should be 1.\n", id); **/
        return 1;
    }
    /** printf("TYPE=\"%.4s\".\n",fileptr+2); **/
    index_offset = READ_BE_DWORD(fileptr + 10);

    while (index_offset <= (filelen - 12))
    {
        id = READ_BE_WORD(fileptr + index_offset);
        data_len = READ_BE_DWORD(fileptr + index_offset + 2);
        offset = READ_BE_DWORD(fileptr + index_offset + 6);
        ext_count = READ_BE_WORD(fileptr + index_offset+10);
    /*    printf("ID: %04d Offset: %ld  Length: %ld Extra: %d\n",
            id, offset, data_len, ext_count); */
        if ((offset > filelen) || ((offset+data_len) > index_offset))
        {
            return -1; /* Not a V1 File */
        }
        if (callback) {
            res = callback(fileptr+2,id,fileptr+offset, data_len);
            if (res < 0) return res;
        }
#if 0
        printf("---------\n");
        hexdump(fileptr+offset, (data_len>128)?128:data_len);
        printf("----------\n");
#endif
        index_offset += (ext_count + 12);
    }
    return 1;
}




int pass1_callback(char * filename, void * data)
{
    unsigned char * fileptr;
    int     filelen;
    int     res;
    char    * pfile;

#ifdef _MSC_VER
    pfile = strrchr(filename, '\\');
#else
    pfile = strrchr(filename, '/');
#endif
    if (!pfile) pfile = filename;
    else pfile++;

    if (strlen(pfile) < 3) return 0; /* Too small, probably a directory */
    if (strcasecmp(pfile,"RSRC.INF") == 0)
    {
        filelen = read_whole_file(filename, &fileptr);

        if ((fileptr) && (filelen > 0))
        {
            meta_data = fileptr;
            meta_len = filelen;
        }
        else
        {
            fprintf(stderr,"Unable to open/read \"RSRC.INF\" !\n");
            return -1;
        }
    }
    else if (strcasecmp(pfile,"DATA.FRK") == 0)
    {
        ;
    } 
    else 
    {
        filelen = read_whole_file(filename, &fileptr);
        if (filelen <= 0) return filelen; 

        if (fileptr)
        {
            res = handle_v1_file(fileptr, filelen, keycallback);
            if (res < 0) { free(fileptr); return res; }

            /* Only include this if successful */
            num_res_files++;
            free(fileptr);
        }
    }
    return 0;
}

/**
 *  Here the files get written out to disk the sb_dir_ptr gets updated
 */
int pass2_callback(char * filename, void * data)
{
    unsigned char * fileptr;
    int     filelen;
    int     res;
    char    * pfile;
    char    type[4];
    int     written;

#ifdef _MSC_VER
    pfile = strrchr(filename, '\\');
#else
    pfile = strrchr(filename, '/');
#endif
    if (!pfile) pfile = filename;
    else pfile++;

    if (strlen(pfile) < 3) return 0; /* Too small, probably a directory */
    if (strcasecmp(pfile,"RSRC.INF") == 0)
    {
        return 0;
    }
    
    filelen = read_whole_file(filename, &fileptr);
    if (filelen <= 0) return filelen; 

    if (!fileptr)
    {
        return 0;
    }

    memset(sb_dir_ptr, 0, 20);
    if (strcasecmp(pfile,"DATA.FRK") == 0)
    {
        des_key_schedule ks;
        int     i;
        char    tmp[8];

        des_set_key((des_cblock *)book_key, ks);

        /** The last block is UNENCRYPTED in this case! **/
        for (i = 0; i < (filelen-7); i+= 8)
        {
            des_ecb_encrypt( (des_cblock *)&fileptr[i],
                (des_cblock *)&fileptr[i],ks,0);
        }
        memset(type,' ', 4);
        memset(sb_dir_ptr, ' ', 4);
    } 
    else 
    {
        res = handle_v1_file(fileptr, filelen, NULL);
        if (res < 0) { free(fileptr); return res; }
        memcpy(sb_dir_ptr, pfile, 4);
        memcpy(type, fileptr + 2, 4);
    }
    /* Don't need any key files no more */
    if (memcmp(type,"!!ky", 4) == 0) 
    {
            free(fileptr);
            return 0;
    }
    processed_res_files++;

    /* +4 is unknown, usually 0 */
    WRITE_BE_DWORD(sb_dir_ptr + 8, filelen); 
    memcpy(sb_dir_ptr+12, type, 4);
    /* +16 is unknown, usually 0 */

    written = fwrite( sb_dir_ptr, 1, 20, fImp);
    if (written != 20) 
    { 
        fprintf(stderr,"Failure writing internal header.\n");
        free(fileptr);
        return -1;
    }
    written = fwrite(fileptr, 1, filelen, fImp);
    if (written != filelen)
    {
        fprintf(stderr,"Failure writing file data.\n");
        free(fileptr);
        return -1;
    }
    free(fileptr);
    sb_dir_ptr += 20;

    if (processed_res_files > num_res_files)
    {
        fprintf(stderr,"Now too many res files (%d vs %d)!  Dying!\n",
            processed_res_files, num_res_files);
        return -1;
    }
    return 0;
}

void decrypt_bookkey()
{
    des_key_schedule ks;

    des_set_key((des_cblock *)devicekey, ks);
    des_ecb_encrypt( (des_cblock *)&book_key[0],
        (des_cblock *)&book_key[0],ks,0);

#if 0 
    printf("Bookkey:\n");
    hexdump(book_key,8);
    printf("-----------\n");
#endif
}

static char * sb_name = "SoftbookL";
int decrypt_check()
{
    unsigned char tmp[16], sb_key[8];
    unsigned long int check;
    int i;
    des_key_schedule ks;

    /* Method 1..
     * 1. Decrypt a block of 0's with "SoftbookL" and use that to decrypt
     *    the check data, 
     * 2. Then treat as four big endian integers, and make sure the
     *    first 3 add up to the 3rd
     */
    memcpy(tmp, check_block, sizeof(tmp));
    memset(sb_key, 0, sizeof(sb_key));
    des_set_key((des_cblock *)sb_name, ks);
    des_ecb_encrypt( (des_cblock *)&sb_key[0],
        (des_cblock *)&sb_key[0],ks,0);
    des_set_key((des_cblock *)sb_key, ks);
    des_ecb_encrypt( (des_cblock *)&tmp[0],
        (des_cblock *)&tmp[0],ks,0);
    des_ecb_encrypt( (des_cblock *)&tmp[8],
        (des_cblock *)&tmp[8],ks,0);
#if 0
    printf("Decrypted check 1 is \n");
    hexdump(tmp, 16);
    printf("---------\n");
#endif

    check = 0;
    for (i = 0; i < 12; i+=4)
    {
        check += READ_BE_DWORD(&tmp[i]);
    }
    if (check != READ_BE_DWORD(&tmp[12])) 
    {
#if 0
        printf("Mismatch: %08lx, should be %08lx\n",
            check, READ_BE_DWORD(&tmp[12]));
#endif
    }
    else return 0;
    

    /* Try method 2  (Bookkey)
     * 1. Decrypt the data block with the bookkey
     * 2. Then treat as four big endian integers, and make sure the
     *    first 3 add up to the 3rd
     */
    memcpy(tmp, check_block, sizeof(tmp));
    des_set_key((des_cblock *)book_key, ks);
    des_ecb_encrypt( (des_cblock *)&tmp[0],
        (des_cblock *)&tmp[0],ks,0);
    des_ecb_encrypt( (des_cblock *)&tmp[8],
        (des_cblock *)&tmp[8],ks,0);
#if 0
    printf("Decrypted check 2 is \n");
    hexdump(tmp, 16);
    printf("----\n");
#endif
   
    check = 0;
    for (i = 0; i < 12; i+=4)
    {
        check += READ_BE_DWORD(&tmp[i]);
    }
    if (check != READ_BE_DWORD(&tmp[12])) 
    {
#if 0
        printf("**MISMATCH**: %08lx, should be %08lx\n",
            check, READ_BE_DWORD(&tmp[12]));
#endif
    }
    else
    {
#if 0
        printf("**GOOD**: %08lx matches %08lx\n",
            check, READ_BE_DWORD(&tmp[12]));
#endif
        return 0;
    }
    return 1;

}

int create_imp_callback(char * filename, void * data)
{
    unsigned char * fileptr;
    int     filelen;
    int     res;

    filelen = read_whole_file(filename, &fileptr);
    if (filelen <= 0) return filelen; 

    if (fileptr)
    {
        res = handle_v1_file(fileptr, filelen, keycallback);
        if (res < 0) return res;
    }
    free(fileptr);
    return 0;
}

char * copy_and_clean(char * dst, char * src, int length)
{
    if (!length) return dst;
    length--;
    while ((*src) && (length > 0))
    {
        if (*src <= ' ') *(dst++) = ' ';
        else
        switch (*src)
        {
        /* Others here? */
        case '/':
        case '\\':
        case '?':
        case ':':
            *(dst++) = ' ';
            break;
        default:
            *(dst++) = *src;
        }
        src++;
        length--;
    }
    *(dst) = '\0';
    return dst;
}

int make_imp_filename( char * filename, int length_in )
{
    char * p, * s;
    int len, j, zeros;
    char * title, * first, * middle, *last;
    int namelen;

    namelen = length_in - 4;

    if (!meta_data) return -1;

    title = last = NULL;
    p = meta_data + 0x24;
    len = meta_len - 0x24;
    zeros = 0;
    for (j = 0; j < len; j++) {
        if (p[j] == 0) {
            zeros++;
            if (zeros == 3) { title = &p[j+1];}
            if (zeros == 4) first = &p[j+1];
            if (zeros == 5) middle = &p[j+1];
            if (zeros == 6) { last = &p[j+1]; break; }
        }
    }
    
    if ((j == len) || (!title) || (!last) || (!*title)) return -1;
    do 
    {
        s = filename;
        if (*first)
        {
            s = copy_and_clean(s,first, namelen);
            namelen -= strlen(first);
            if (namelen < 0) break;
        }
        if (s != filename)
        {
            namelen--;
            if (namelen < 0) break;
            *(s++) = ' ';
        }
        if (*middle)
        {
            s = copy_and_clean(s,middle, namelen);
            namelen -= strlen(middle);
            if (namelen < 0) break;
        }
        if (s != filename)
        {
            namelen--;
            if (namelen < 0) break;
            *(s++) = ' ';
        }
        if (*last)
        {
            s = copy_and_clean(s,last, namelen);
            namelen -= strlen(last);
            if (namelen < 0) break;
        }
        if (s != filename)
        {
            namelen--;
            if (namelen < 0) break;
            *(s++) = '_';
        }
        s = copy_and_clean(s,title, namelen);
    } while (0);
    if (namelen < 0) 
    {
        s = copy_and_clean(filename, title, length_in - 4);
    }
    strcat(filename,".imp");
    return 0;
}

int res2imp(char * directory, byte * in_devicekey)
{
    int err;
    char * pDir; 
    char impfilename[FILENAME_MAX];
    char resname[FILENAME_MAX];
#ifdef _MSC_VER
    char sep = '\\';
#else
    char sep = '/';
#endif


    meta_data = sb_header = NULL;

    if (in_devicekey) {
    	memcpy(devicekey, in_devicekey,8);
    	has_devicekey++;
    }


    pDir = strrchr(directory,sep);
    if (!pDir) pDir = directory;
    else pDir++;
    if (!*pDir)
    {
        pDir--;
        while (pDir > directory)
        {
            if (*(pDir-1) == sep) break;
            pDir--;
        }
        strncpy(resname, pDir, strlen(pDir) - 1);
    }
    else
    {
        strcpy(resname, pDir);
    }
    pDir = &resname[0];

    num_res_files = 1;  /* Must have "DATA.FRK" */
    meta_data = NULL;
    meta_len = 0;
    err = foreach_file(directory, pass1_callback, NULL);
    if (err)
    {
        fprintf(stderr,"Error in reading pass for \"%s\" \n.", directory);
        err = -100;
        goto error;
    }
    if (!meta_data) 
    {
        fprintf(stderr,"Error in \"%s\" - No RSRC.INF data found!  Can't convert\n", directory);
        err = -100;
        goto error;
    }
    if (need_aes_key) {
        char * s;

        s = astrcat("AEBK:",meta_data+0x24);
        if (!s) goto error;
        if (get_book_key_from_prefs(s,book_key) != 0) { free(s); 
            err = -200; goto error;}
        free(s);
    }

    if (found_keydata)
    {
        if (!has_devicekey)
        {
            fprintf(stderr,"Error in \"%s\" - Encrypted file and no device key!\n", directory);
            err = -200;
            goto error;
        } 
        if (need_aes_key == 0) decrypt_bookkey();	
        if (decrypt_check() != 0) 
        {
            fprintf(stderr,"Error in \"%s\" - Key appears to be wrong..\n", directory);
            err = -200;
            goto error;
        }
        /* Not going to put the !!ky in the .IMP file */
        num_res_files--;
    }

    err = make_imp_filename(impfilename, sizeof(impfilename));
    if (err < 0) 
    { 
        fprintf(stderr,"Error with \"%s\" - Unable to create a .imp filename\n", directory);
        err = -300;
        goto error;
    }
    printf("Writing \"%s\" to file: %s\n", directory, impfilename);

    fImp = fopen(impfilename,"wb");
    if (!fImp)
    {
        fprintf(stderr,"Fatal error opening \"%s\"! \n", impfilename);
        err = -300;
        goto error;
    }
    sb_header_len = 48 + (meta_len - 36) + strlen(pDir) + (num_res_files * 20);
    sb_header = (unsigned char *)malloc(sb_header_len);
    if (!sb_header)
    {
        fprintf(stderr,"Can't allocate header!\n");
        fclose(fImp);
        unlink(impfilename);
        err = -300;
        goto error;

    }
    memset(sb_header, 0, sb_header_len);
    WRITE_BE_WORD(&sb_header[0],2);
    memcpy(&sb_header[2],"BOOKDOUG",8);
    WRITE_BE_WORD(&sb_header[0x12], num_res_files);
    WRITE_BE_WORD(&sb_header[0x14], strlen(pDir)); 
    /* The remaining length in the header from this point on */
    WRITE_BE_WORD(&sb_header[0x16], (48 + (meta_len - 36)) - 0x18);
    /* 12 -- this is the BOOKDOUG header.. this DOES overlap the earlier 36 */
    memcpy(&sb_header[0x18], meta_data + 12, meta_len - 12);
    strcpy(&sb_header[48 + (meta_len - 36)],pDir);

    /* should clean out the encryption flag */
    WRITE_BE_DWORD(&sb_header[0x24], 0);
    
    sb_dir_ptr = sb_header + (48 + (meta_len - 36) + strlen(pDir));
    
    /* skip over the space for the header... */
    (void)fseek(fImp, sb_header_len, SEEK_SET);

    /* Now, write out all the files */
    processed_res_files = 0;
    err = foreach_file(directory, pass2_callback, NULL);
    if (err)
    {
        fprintf(stderr,"Error in writing pass.\n");
        fclose(fImp);
        unlink(impfilename);
        err = -300;
        goto error;
    }
    fseek(fImp, 0, SEEK_SET);
    err = fwrite(sb_header, 1, sb_header_len, fImp);
    if (err != sb_header_len)
    {
        fprintf(stderr,"Error writing the main header.\n");
        fclose(fImp);
        unlink(impfilename);
        err = -300;
        goto error;
    }
    return 0;
error:
    if (meta_data != NULL) free(meta_data);
    if (sb_header != NULL) free(sb_header);
    return err;
}
